package edu.kit.ipd;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.ShortBuffer;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.opengl.GLSurfaceView;
import android.opengl.GLUtils;

public class OpenGLView extends GLSurfaceView {
	
	private Renderer renderer = new Renderer();
	
	private static String glErrorToString(int error) {
		switch (error) {
		case GL10.GL_NO_ERROR:          return "no error";
		case GL10.GL_INVALID_ENUM:      return "invalid enum";
		case GL10.GL_INVALID_OPERATION: return "invalid operation";
		case GL10.GL_INVALID_VALUE:     return "invalid value";
		case GL10.GL_STACK_OVERFLOW:    return "stack overflow";
		case GL10.GL_STACK_UNDERFLOW:   return "stack underflow";
		case GL10.GL_OUT_OF_MEMORY:     return "out of memory";
		default:                        return "unknown error";
		}
	}
	
	private static void checkGlState(GL10 gl) {
		int error = gl.glGetError();
		if (error != GL10.GL_NO_ERROR) {
			System.err.println("OpenGL Error: " + glErrorToString(error));
			throw new RuntimeException("OpenGL Error: " + glErrorToString(error));
		}
	}
	
	private static void printGLInfo(GL10 gl) {
		System.err.printf("Vendor: %s\n", gl.glGetString(GL10.GL_VENDOR));
		System.err.printf("Renderer: %s\n", gl.glGetString(GL10.GL_RENDERER));
		System.err.printf("Version: %s\n", gl.glGetString(GL10.GL_VERSION));
		System.err.printf("Extensions: %s\n", gl.glGetString(GL10.GL_EXTENSIONS));
	}
	
	private static class Renderer implements GLSurfaceView.Renderer {
		static final FloatBuffer cubeVertexBuffer;
		static final ShortBuffer trianglesBuffer;
		static final FloatBuffer uvCoords;
		static FloatBuffer sphereVertices;
		static ShortBuffer sphereTriangles;
		static FloatBuffer sphereTexCoords;
		
		private static FloatBuffer makeDirectFloatBuffer(float[] array) {
			ByteBuffer storage = ByteBuffer.allocateDirect(array.length * (Float.SIZE/8));
			storage.order(ByteOrder.nativeOrder());
			FloatBuffer buffer = storage.asFloatBuffer();
			buffer.put(array);
			buffer.position(0);
			return buffer;
		}
		
		private static ShortBuffer makeDirectShortBuffer(short[] array) {
			ByteBuffer storage = ByteBuffer.allocateDirect(array.length * (Short.SIZE/8));
			storage.order(ByteOrder.nativeOrder());
			ShortBuffer buffer = storage.asShortBuffer();
			buffer.put(array);
			buffer.position(0);
			return buffer;
		}
		
		/**
		 * Create geometry and texture coordinates for a sphere
		 * (simple and naive version)
		 */
		private static void makeSphereTriangles() {
			int L = 13;
			int P = 13;
			int nverts = L*P;
			int ntris  = (L-1) * (P-1) * 2;
			
			float[] verts = new float[nverts * 3];
			float[] texCoords = new float[nverts * 2];
			short[] tris = new short[ntris * 3];
			int vptr = 0;
			int tcptr = 0;
			int triptr = 0;
			for (int l = 0; l < L; ++l) {
				double rad = Math.sin(Math.PI * ((double) l / ((double) L-1)));
				double y   = - Math.cos(Math.PI * ((double) l / ((double) L-1)));
				double v   = 1.0 - l / ((double)L-1);
				for (int p = 0; p < P; ++p) {
					double x = Math.sin( (2*Math.PI*p)/((double) P-1)) * rad;
					double z = Math.cos( (2*Math.PI*p)/((double) P-1)) * rad;
					double u = (double) p / ((double)P-1);
					verts[vptr++] = (float) x;
					verts[vptr++] = (float) y;
					verts[vptr++] = (float) z;
					texCoords[tcptr++] = (float) u;
					texCoords[tcptr++] = (float) v;
					
					/* create triangles between vertices */
					if (l > 0 && p < P-1) {
						tris[triptr++] = (short) ((l-1)*P + p);
						tris[triptr++] = (short) ((l-1)*P + p + 1);
						tris[triptr++] = (short) (l*P + p);
					}
					if (l > 0 && p > 0) {
						tris[triptr++] = (short) ((l-1)*P + p);
						tris[triptr++] = (short) (l*P + p);
						tris[triptr++] = (short) (l*P-1 + p);						
					}
				}
			}
			sphereVertices = makeDirectFloatBuffer(verts);
			sphereTexCoords = makeDirectFloatBuffer(texCoords);
			sphereTriangles = makeDirectShortBuffer(tris);
		}
		
		static {
			float[] cubeVertices = new float[] {
				-1, -1, -1,
				-1,  1, -1,
				 1,  1, -1,
				 1, -1, -1,
				-1, -1,  1,
				-1,  1,  1,
				 1,  1,  1,
				 1, -1,  1
			};
			cubeVertexBuffer = makeDirectFloatBuffer(cubeVertices);
			
			float[] uvcoords = new float[] {
				 0,  0,
				 0,  1,
				 1,  1,
				 1,  0,
				 0,  0,
				 0,  1,
				 1,  1,
				 1,  0,
			};
			uvCoords = makeDirectFloatBuffer(uvcoords);
			
			short[] triangles = new short[] {
				/* front */
				0, 1, 2,
				2, 3, 0,
				/* back */
				6, 5, 4,
				4, 7, 6,
				/* top */
				4, 5, 1,
				1, 0, 4,
				/* bottom */
				2, 6, 7,
				7, 3, 2,
				/* left */
				7, 4, 0,
				0, 3, 7,
				/* right */
				1, 5, 6,
				6, 2, 1
			};
			trianglesBuffer = makeDirectShortBuffer(triangles);
			
			makeSphereTriangles();
		}
		
		private float angle = 0;
		
		private void drawSphere(GL10 gl) {
			gl.glEnableClientState(GL10.GL_VERTEX_ARRAY);
			gl.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY);
			
			gl.glTexCoordPointer(2, GL10.GL_FLOAT, 0, sphereTexCoords);
			gl.glVertexPointer(3, GL10.GL_FLOAT, 0, sphereVertices);
			gl.glDrawElements(GL10.GL_TRIANGLES, sphereTriangles.capacity(), GL10.GL_UNSIGNED_SHORT, sphereTriangles);			
		}
			
		@SuppressWarnings("unused")
		private void drawCube(GL10 gl) {
			gl.glEnableClientState(GL10.GL_VERTEX_ARRAY);
			gl.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY);			
			
			gl.glBindTexture(GL10.GL_TEXTURE_2D, earthTexId);
			gl.glTexCoordPointer(2, GL10.GL_FLOAT, 0, uvCoords);
			gl.glVertexPointer(3, GL10.GL_FLOAT, 0, cubeVertexBuffer);
			gl.glDrawElements(GL10.GL_TRIANGLES, trianglesBuffer.capacity(), GL10.GL_UNSIGNED_SHORT, trianglesBuffer);			
		}
		
		private void texImage2D(GL10 gl, Bitmap bitmap, int mipmapLevel) {
			assert(bitmap.getConfig() == Bitmap.Config.ARGB_8888);
			ByteBuffer bytebuf = ByteBuffer.allocateDirect(bitmap.getHeight() * bitmap.getWidth() * 4);
			bytebuf.order(ByteOrder.nativeOrder());
			IntBuffer pixelbuf = bytebuf.asIntBuffer();
			 
			for(int y = 0; y < bitmap.getHeight(); y++)
			  for(int x = 0; x < bitmap.getWidth(); x++) {
			    pixelbuf.put(bitmap.getPixel(x, y));
			  }
			pixelbuf.position(0);
			bytebuf.position(0);
			 
			gl.glTexImage2D(GL10.GL_TEXTURE_2D, mipmapLevel, GL10.GL_RGBA, bitmap.getWidth(), bitmap.getHeight(), 0, GL10.GL_RGBA, GL10.GL_UNSIGNED_BYTE, pixelbuf);
		}
		
		/** Erzeugt OpenGL Textur aus einer Bitmap
		 * @return Neu erzeugte TexturID */
		private int loadGlTexture(GL10 gl, Bitmap bitmap) {
			int[] tids = new int[1];
			
			gl.glEnable(GL10.GL_TEXTURE_2D);
			gl.glGenTextures(1, tids, 0);

			gl.glBindTexture(GL10.GL_TEXTURE_2D, tids[0]);
			
			/* check that texture width and high is a power of 2 */ 
			assert (bitmap.getWidth() & (bitmap.getWidth()-1)) == 0;
			assert (bitmap.getHeight() & (bitmap.getHeight()-1)) == 0;
			GLUtils.texImage2D(GL10.GL_TEXTURE_2D, 0, bitmap, 0);
			gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MIN_FILTER, GL10.GL_LINEAR);
			gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MAG_FILTER, GL10.GL_LINEAR);
			
			return tids[0];
		}
		
		private int loadGlTextureAsset(GL10 gl, Context context, String name) {
			int res = 0;
			try {
				Bitmap bitmap = BitmapFactory.decodeStream(context.getAssets().open(name));
				res = loadGlTexture(gl, bitmap);
			} catch (IOException e) {
				/* bad style */
				throw new RuntimeException(e);
			}
			return res;
		}
		
		private void drawSun(GL10 gl) {
			gl.glBindTexture(GL10.GL_TEXTURE_2D, sunTexId);
			drawSphere(gl);
		}
		
		private void drawEarth(GL10 gl) {
			gl.glBindTexture(GL10.GL_TEXTURE_2D, earthTexId);
			drawSphere(gl);
		}
		
		private void drawMars(GL10 gl) {
			gl.glBindTexture(GL10.GL_TEXTURE_2D, marsTexId);
			drawSphere(gl);
		}
		
		public void onDrawFrame(GL10 gl) {
			gl.glClearColor(0, 0, 0, 1);
			gl.glClear(GL10.GL_COLOR_BUFFER_BIT | GL10.GL_DEPTH_BUFFER_BIT);
			
			/* this should really be done with a timer... but for demonstration purposes this is ok */
			angle += 1;
			
			gl.glMatrixMode(GL10.GL_MODELVIEW);
			gl.glLoadIdentity();
			gl.glTranslatef(0, 0, -5);
			
			drawSun(gl);
			
			gl.glPushMatrix();
			gl.glRotatef(angle, 0, 1, 0);
			gl.glTranslatef(2f, 0, 0);
			gl.glScalef(0.6f, 0.6f, 0.6f);
			gl.glRotatef(23f, 0, 0, 1);
			gl.glRotatef(angle*2, 0, 1, 0);
			drawEarth(gl);
			gl.glPopMatrix();
			
			gl.glPushMatrix();
			gl.glRotatef(angle*1.6f, 0, 1, 0);
			gl.glTranslatef(0, 0, 3f);
			gl.glScalef(0.3f, 0.3f, 0.3f);
			drawMars(gl);
			gl.glPopMatrix();
			
			checkGlState(gl);
		}

		public void onSurfaceChanged(GL10 gl, int width, int height) {
			/* setup viewport */
			gl.glViewport(0, 0, width, height);
			
			/* setup projection matrix */
			gl.glMatrixMode(GL10.GL_PROJECTION);
			gl.glLoadIdentity();
			float ratio = (float) width / height;
			gl.glFrustumf(-ratio, ratio, -1, 1, 1, 10);
			
			checkGlState(gl);
		}
		
		public Context context;
		private int earthTexId;
		private int sunTexId;
		@SuppressWarnings("unused")
		private int moonTexId;
		private int marsTexId;

		public void onSurfaceCreated(GL10 gl, EGLConfig config) {
			printGLInfo(gl);
			
			gl.glEnable(GL10.GL_DEPTH_TEST);
			gl.glEnable(GL10.GL_CULL_FACE);
			gl.glDisable(GL10.GL_BLEND);
			gl.glShadeModel(GL10.GL_SMOOTH);
			gl.glDisable(GL10.GL_DITHER);
			gl.glHint(GL10.GL_PERSPECTIVE_CORRECTION_HINT, GL10.GL_FASTEST);
			
			earthTexId = loadGlTextureAsset(gl, context, "earth_small.jpg");
			sunTexId = loadGlTextureAsset(gl, context, "sun_small.jpg");
			moonTexId = loadGlTextureAsset(gl, context, "moon_small.jpg");
			marsTexId = loadGlTextureAsset(gl, context, "mars_small.jpg");
			
			checkGlState(gl);
		}
	}

	public OpenGLView(Context context) {
		super(context);
		renderer.context = context;
		setRenderer(renderer);
	}

}
